package invalid.thexbot;

import java.io.FileInputStream;
import invalid.thexbot.net.LinePoller;
import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Properties;
import java.util.logging.Level;
import invalid.thexbot.exceptions.ConfigurationException;
import invalid.thexbot.exceptions.ParserException;
import invalid.thexbot.irc.EntityManager;
import invalid.thexbot.protocol.DarenetProtocol;
import invalid.thexbot.protocol.Message;
import invalid.thexbot.protocol.Protocol;
import invalid.thexbot.scripting.ModuleManager;

/**
 * Main entry point for the X bot system.
 */
public final class TheXBot {
	public enum XState {
		BOOTING, BURSTING, RUNNING, SHUTTING_DOWN;
	}
	private static final Logger log = Logger.getLogger(TheXBot.class.getName());
	public static ConfigurationManager conf;
	public static final int start_time;
	public static boolean bursting;
	private static boolean shutdown = false;
	private static EntityManager em;
	private static ModuleManager mm;
	private static XState state = XState.BOOTING;

	static {
		start_time = Util.getTimestamp();
	}

	/**
	 * Main entry point of the X bot. Creates a new Protocol instance and connects.
	 *
	 * @param args the command line arguments
	 */
	public static void main(String[] args)
	{
		log.log(Level.INFO, "Starting X...");

		log.log(Level.INFO, "Loading configuration...");
		final Properties props = new Properties();
		try {
			props.load(new FileInputStream("x.properties"));
		} catch (IOException e) {
			log.log(Level.SEVERE, "Unable to open x.properties.", e);
			return;
		}

		try {
			conf = new ConfigurationManager(props);
		} catch (ConfigurationException e) {
			log.log(Level.SEVERE, null, e);
			return;
		}
		log.log(Level.INFO, "Configuration checked cleanly.");

		if (conf.isDebug()) {
			log.setLevel(Level.ALL);
		}

		Protocol p;
		LinePoller lp;
		Socket s;
		try {
			log.log(Level.WARNING, "Connecting to uplink...");
			s = new Socket("127.0.0.1", 7326);

			p = new DarenetProtocol(s);
			lp = new LinePoller(s);
		} catch (UnknownHostException e) {
			log.log(Level.SEVERE, "Could not resolve uplink host. Aborting.", e);
			return;
		} catch (IOException e) {
			log.log(Level.SEVERE, "Error connecting to uplink.", e);
			return;
		} catch (ConfigurationException e) {
			log.log(Level.SEVERE, "Configuration exception while connecting.", e);
			return;
		}

		em = new EntityManager(p);
		mm = new ModuleManager(p);
		mm.loadModules();
		p.handshake();
		state = XState.BURSTING;
		p.burst();
		lp.start();

		String line;
		while (!shutdown) {
			/* This is a little throttler because the CPU usage would spike due to the
			 * infinite loop here. The speed of sleeping 16 ms is relatively sane and
			 * responsive, but we skip it while there's a server bursting so we don't
			 * get disconnected.
			 */
			if (!bursting) {
				try {
					Thread.sleep(16);
				} catch (InterruptedException e) {
					log.log(Level.FINE, "Main thread interrupted while sleeping.", e);
				}
			}

			line = lp.poll();
			if (line == null) {
				continue;
			}

			try {
				Message m = p.parse(line);
				p.interpret(m);
			} catch (ParserException e) {
				log.log(Level.WARNING, "Unable to parse line: " + line, e);
			} catch (Exception e) {
				log.log(Level.WARNING, "Exception while parsing message: " + line, e);
			}
		}
		try {
			s.shutdownInput();
			s.shutdownOutput();
			s.close();
		} catch (IOException e) {
		}
		mm.cleanup();
		lp.disconnect();
		p.disconnect();
		log.log(Level.INFO, "X finished shutting down.");
	}

	/**
	 * Shuts the X bot down.
	 */
	public static void shutdown()
	{
		shutdown = true;
	}

	/**
	 * @return the {@link EntityManager} running right now
	 */
	public static EntityManager em()
	{
		return em;
	}

	/**
	 * @return the {@link ModuleManager} running right now
	 */
	public static ModuleManager mm()
	{
		return mm;
	}
	
	/**
	 * @return the current state
	 */
	public static XState getState()
	{
		return state;
	}
	
	/**
	 * Changes the state.
	 * @param newstate the new state
	 */
	public static void setState(XState newstate)
	{
		state = newstate;
	}
}
