package invalid.thexbot;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.logging.Level;

/**
 * Utility functions for X.
 *
 * @since 0.5.0
 */
public final class Util {
	private static final Logger log = Logger.getLogger(Util.class.getName());

	public static <T> String join(String intermezzo, T[] array)
	{
		if (intermezzo == null || array == null || array.length == 0) {
			return "";
		}

		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < array.length; i++) {
			sb.append(array[i].toString());
			if (i < (array.length - 1)) {
				sb.append(intermezzo);
			}
		}
		
		return sb.toString();
	}

	/**
	 * @return the current UNIX timestamp.
	 * @since 0.5.0
	 */
	public static int getTimestamp()
	{
		return (int)(System.currentTimeMillis() / 1000L);
	}

	/**
	 * Counts the number of times a given character occurs in a given string.
	 *
	 * @param s the string containing the character
	 * @param search the character to be counted
	 * @return the number of times the character occurs
	 * @since 0.5.0
	 */
	public static int countCharacters(final String s, final char search)
	{
		int count = 0;

		for (final char c : s.toCharArray()) {
			if (c == search) {
				count++;
			}
		}

		return count;
	}
	/* P10 utilities, mostly stolen from ircu */
	/**
	 * Number of bits encoded in one numnick character.
	 *
	 * @since 0.5.0
	 */
	private static final int NUMNICKLOG = 6;
	/**
	 * Bitmask to select value of next numnick character.
	 *
	 * @since 0.5.0
	 */
	private static final int NUMNICKMASK = 63;          /* (NUMNICKBASE-1) */

	/**
	 * Number of clients representable in a numnick.
	 *
	 * @since 0.5.0
	 */
	private static final int NN_MAX_CLIENT = 262144;    /* (NUMNICKBASE ^ 3 */

	/**
	 * Converts a numeric to the corresponding character. The following characters are currently
	 * known to be forbidden:
	 *
	 * '\\0' : Because we use '\\0' as end of line.
	 *
	 * ' ' : Because parse_*() uses this as parameter separator.
	 *
	 * ':' : Because parse_server() uses this to detect if a prefix is a numeric or a name.
	 *
	 * '+' : Because m_nick() uses this to determine if parv[6] is a umode or not.
	 *
	 * '&', '#', '$', '
	 *
	 * @' and '%' : Because m_message() matches these characters to detect special cases.
	 * @since 0.5.0
	 */
	private static final char[] convert2y = {
		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
		'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
		'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
		'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '[', ']'
	};
	/**
	 * Converts a character to its (base64) numnick value.
	 *
	 * @since 0.5.0
	 */
	static final char convert2n[] = {
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 0, 0, 0, 0, 0, 0,
		0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
		15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 62, 0, 63, 0, 0,
		0, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
		41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};

	/**
	 * @param count the number of clients
	 * @return true if the given client count is the maximum number of clients we can encode as
	 * numeric nick.
	 * @since 0.5.0
	 */
	public static boolean isMaxClientNumber(final int count)
	{
		return count == NN_MAX_CLIENT;
	}

	/**
	 * Convert a string to its value as a numnick. Returns long because Java doesn't know
	 * unsigned int.
	 *
	 * @param s Numnick string to decode
	 * @return Numeric name value
	 * @since 0.5.0
	 */
	public static long base64toint(final String s)
	{
		long i = 0;
		for (final char c : s.toCharArray()) {
			i <<= NUMNICKLOG;
			i += convert2n[c];
		}

		log.log(Level.FINEST, "Parsed " + s + " to " + i);
		return i;
	}

	/**
	 * Encode a number as a numnick.
	 *
	 * @param val Value to encode.
	 * @param count Number of numnick digits to write.
	 * @return the number as numnick
	 * @since 0.5.0
	 */
	public static String inttobase64(int val, int count)
	{
		char[] buf = new char[count];
		while (count > 0) {
			buf[--count] = convert2y[val & NUMNICKMASK];
			val >>= NUMNICKLOG;
		}
		return new String(buf);
	}

	/**
	 * Decode an IP address from base64.
	 *
	 * @param input Input buffer to decode
	 * @return IP address
	 * @since 0.5.0
	 */
	public static InetAddress base64toip(final String input)
	{
		InetAddress address;
		if (input.length() == 6) {
			long in = base64toint(input);
			try {
				address = InetAddress.getByAddress(new byte[]{(byte)(in >> 24), (byte)((in >> 16) & 255),
									      (byte)((in >> 8) & 255), (byte)(in & 255)});
			} catch (UnknownHostException e) {
				log.log(Level.WARNING, "Invalid IP address " + in + " as " + input);
				return null;
			}
		} else {
			char[] addr = new char[]{0, 0, 0, 0, 0, 0, 0, 0}; // 8 elements

			/* Kept this block from ircu, I don't know how exactly this works and I don't care enough to
			 * make it spit out proper byte[].
			 */
			int pos = 0, idx = 0;
			do {
				if (input.charAt(idx) == '_') {
					int left;
					for (left = (25 - input.length()) / 3; left != 0; left--) {
						addr[pos++] = 0;
					}
					idx++;
				} else {
					char accum = convert2n[input.charAt(idx++)];
					accum = (char)((accum << NUMNICKLOG) | convert2n[input.charAt(idx++)]);
					accum = (char)((accum << NUMNICKLOG) | convert2n[input.charAt(idx++)]);
					addr[pos++] = accum;
				}
			} while (pos < 8);

			byte[] bytes = new byte[16];
			for (int i = 0; i < 16; i++) {
				// upper byte
				if (i == 0 || i % 2 == 0) {
					bytes[i] = (byte)(addr[i / 2] >> 8);
				} // lower byte
				else // Integer division! 1/2 = 0, not 1.
				{
					bytes[i] = (byte)(addr[i / 2] & 255);
				}
			}

			try {
				address = InetAddress.getByAddress(bytes);
			} catch (UnknownHostException e) {
				log.log(Level.WARNING, "Invalid IP address " + input);
				return null;
			}
		}

		log.log(Level.FINEST, "Parsed IP: " + input + " to " + address.getHostAddress());
		return address;
	}
}
