package invalid.thexbot.irc;

import java.util.EnumSet;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Set;

/**
 * Keeps track of an IRC channel.
 *
 * @since 0.5.0
 */
public final class Channel {
	private final String name;
	private int ts;
	private boolean is_persistent;
	// user -> CUS
	private final Map<User, EnumSet<ChannelUserStatus>> users = new IdentityHashMap<>();

	/**
	 * Creates a new channel.
	 *
	 * @param name the name of the channel
	 * @param ts the timestamp of the channel
	 * @since 0.5.0
	 */
	public Channel(String name, int ts)
	{
		this.name = name;
		this.ts = ts;
	}

	/**
	 * @return the name of this channel
	 */
	public String getName()
	{
		return this.name;
	}

	/**
	 * Sets this channel as persistent, i.e., it does not get destroyed when the last user
	 * parts.
	 *
	 * @param is_persistent true if the channel is supposed to be persistent
	 * @since 0.5.0
	 */
	public void setPersistent(boolean is_persistent)
	{
		this.is_persistent = is_persistent;
	}

	/**
	 * @return true if this channel is persistent
	 * @see Channel#setPersistent()
	 * @since 0.5.0
	 */
	public boolean isPersistent()
	{
		return this.is_persistent;
	}

	/**
	 * @return the channel timestamp
	 * @since 0.5.0
	 */
	public int getTS()
	{
		return this.ts;
	}

	/**
	 * Changes the channel timestamp.
	 *
	 * @param ts the timestamp to set
	 * @since 0.5.0
	 */
	public void setTS(int ts)
	{
		this.ts = ts;
	}

	/**
	 * Adds a {@link User} as peon to the channel. This is a short-hand for
	 * {@link Channel#addUser(org.darenet.thexbot.irc.User, java.util.EnumSet)} with an empty
	 * EnumSet.
	 *
	 * @param u the User to add
	 * @since 0.5.0
	 */
	public void addUser(User u)
	{
		this.addUser(u, EnumSet.noneOf(ChannelUserStatus.class));
	}

	/**
	 * Adds a {@link User} with the given {@link ChannelUserStatus} set to the channel. This
	 * does not update the internal state of the User!
	 *
	 * @param u the User to add
	 * @param cus the {@link EnumSet} of ChannelUserStatus flag to add the User with
	 * @since 0.5.0
	 */
	public void addUser(User u, EnumSet<ChannelUserStatus> cus)
	{
		this.users.put(u, cus);
	}

	/**
	 * Removes a {@link User} from a channel. THis does not update the internal state of the
	 * User!
	 *
	 * @param u the User to remove
	 * @since 0.5.0
	 */
	public void removeUser(User u)
	{
		this.users.remove(u);
	}

	/**
	 * @return the list of channel {@link User}s with status
	 * @since 0.5.0
	 */
	public Map<User, EnumSet<ChannelUserStatus>> getUsersWithStatus()
	{
		return this.users;
	}

	/**
	 * @return the channel {@link User}s without status
	 * @since 0.5.0
	 */
	public Set<User> getUsers()
	{
		return this.users.keySet();
	}

	/**
	 * Finds the status of a given {@link User} in this channel.
	 *
	 * @param u the User to look for
	 * @return the {@link EnumSet} of status modes (possibly empty) or null if the user is not
	 * on the channel.
	 */
	public EnumSet<ChannelUserStatus> getStatus(final User u)
	{
		return this.users.get(u);
	}
	
	/**
	 * Cleans out all lists.
	 */
	public void destroy()
	{
		this.users.clear();
	}
}
