package invalid.thexbot.irc;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import invalid.thexbot.TheXBot;
import invalid.thexbot.protocol.Protocol;

/**
 * Keeps track of various IRC entities, such as servers, users and channels.
 *
 * @since 0.5.0
 */
public final class EntityManager {
	private final Map<String, Server> servers_by_id = new HashMap<>();
	// lowercase name -> Server
	private final Map<String, Server> servers_by_name = new HashMap<>();
	private final Map<String, User> users_by_id = new HashMap<>();
	// lowercase nick -> User
	private final Map<String, User> users_by_nick = new HashMap<>();
	// lowercase channel name -> Channel
	private final Map<String, Channel> channels = new HashMap<>();
	// numeric -> PseudoClient
	private final Map<String, PseudoClient> pseudoclients_by_id = new HashMap<>();
	private final Protocol protocol;

	/**
	 * Initializes this entity manager.
	 *
	 * @param protocol the {@link Protocol} to use. Note that EntityManager does not actually
	 * send anything over the network using the Protocol; it is only required because IRCds have
	 * varying case mapping rules.
	 * @since 0.5.0
	 */
	public EntityManager(Protocol protocol)
	{
		this.protocol = protocol;
	}

	/**
	 * Adds a {@link Server} to the server list.
	 *
	 * @param s the server to add
	 * @since 0.5.0
	 */
	public void addServer(final Server s)
	{
		// TS5 jupes may not have a numeric
		if (s.getNumeric() != null) {
			servers_by_id.put(s.getNumeric(), s);
		}
		servers_by_name.put(s.getName().toLowerCase(), s);
	}

	/**
	 * Finds a server by name.
	 *
	 * @param name the name to look for
	 * @return the {@link Server} in question or null if not found
	 * @since 0.5.0
	 */
	public Server findServerByName(final String name)
	{
		return servers_by_name.get(name.toLowerCase());
	}

	/**
	 * Finds a server by numeric. The numeric must first be converted to the usual String
	 * representation if the protocol uses encoded numbers.
	 *
	 * @param numeric the numeric to look for
	 * @return the {@link Server} in question or null if not found
	 * @since 0.5.0
	 */
	public Server findServerByNumeric(final String numeric)
	{
		return servers_by_id.get(numeric);
	}

	/**
	 * Finds a server by either name or numeric. Note that lookup by numeric is performed first
	 * so if you know in advance which lookup you need, use more specialized functions.
	 *
	 * @param source the name or numeric
	 * @return the {@link Server} in question or null if not found
	 * @since 0.5.0
	 */
	public Server findServer(final String source)
	{
		Server s;
		if ((s = findServerByNumeric(source)) != null) {
			return s;
		}

		return findServerByName(source);
	}

	/**
	 * Removes a {@link Server} from the internal list.
	 *
	 * @param s the Server to remove
	 * @param include_downlinks if true, all servers linked to the one to be removed will also
	 * be purged from the list; needs to be set to false for some protocols where the uplink
	 * generates the list of servers that were affected by a split and sends that around
	 */
	public void removeServer(Server s, boolean include_downlinks)
	{
		if (include_downlinks) {
			// clean out downlinks first
			for (final Server downlink : s.getDownlinks()) {
				removeServer(downlink, true);
			}
		}

		// outermost child node here
		this.servers_by_id.remove(s.getNumeric());
		this.servers_by_name.remove(s.getName().toLowerCase());
		for (final User u : s.getUsers()) {
			removeUser(u);
		}
	}

	/**
	 * Adds a user to the user list.
	 *
	 * @param u the user to add
	 * @since 0.5.0
	 */
	public void addUser(final User u)
	{
		users_by_nick.put(protocol.stringToLowerCase(u.getNick()), u);
		users_by_id.put(u.getNumeric(), u);
	}

	/**
	 * Finds a user by nick.
	 *
	 * @param nick the nick to look for
	 * @return the {@link User} in question or null if not found
	 * @since 0.5.0
	 */
	public User findUserByNick(final String name)
	{
		return users_by_nick.get(protocol.stringToLowerCase(name));
	}

	/**
	 * Finds a user by numeric. The numeric must first be converted to the usual String
	 * representation if the protocol uses encoded numbers.
	 *
	 * @param numeric the numeric to look for
	 * @return the {@link User} in question or null if not found
	 * @since 0.5.0
	 */
	public User findUserByNumeric(final String numeric)
	{
		return users_by_id.get(numeric);
	}

	/**
	 * Finds a user by either nick or numeric. Note that lookup by numeric is performed first so
	 * if you know in advance which lookup you need, use more specialized functions.
	 *
	 * @param source the nick or numeric
	 * @return the {@link User} in question or null if not found
	 * @since 0.5.0
	 */
	public User findUser(final String source)
	{
		User u;
		if ((u = findUserByNumeric(source)) != null) {
			return u;
		}

		return findUserByNick(source);
	}

	/**
	 * Removes a {@link User} from all records and updates {@link Channel} state.
	 *
	 * @param u the User to remove
	 * @since 0.5.0
	 */
	public void removeUser(final User u)
	{
		for (final Channel c : u.getChannels()) {
			c.removeUser(u);
			if (c.getUsers().isEmpty() && !c.isPersistent()) {
				// fuck zannels
				removeChannel(c);
			}
		}

		u.getServer().removeUser(u);
		u.quit();
		this.users_by_id.remove(u.getNumeric());
		this.users_by_nick.remove(this.protocol.stringToLowerCase(u.getNick()));
	}

	/**
	 * Adds a {@link PseudoClient} to the list. Do not call
	 * {@link EntityManager#addUser(org.darenet.thexbot.irc.User)} separately.
	 *
	 * @param u the PseudoClient to add
	 * @since 0.5.0
	 */
	public void addPseudoClient(final PseudoClient u)
	{
		addUser(u);
		this.pseudoclients_by_id.put(u.getNumeric(), u);
		if (TheXBot.getState() == TheXBot.XState.RUNNING) {
			this.protocol.introducePseudoClient(u);
		}
	}

	/**
	 * Adds a {@link PseudoClient} to the list. Do not call
	 * {@link EntityManager#removeUser(org.darenet.thexbot.irc.User)} separately.
	 *
	 * @param u the PseudoClient to add
	 * @since 0.5.0
	 */
	public void removePseudoClient(final PseudoClient u)
	{
		removeUser(u);
		this.pseudoclients_by_id.remove(u.getNumeric());
		if (TheXBot.getState() == TheXBot.XState.BURSTING
			|| TheXBot.getState() == TheXBot.XState.RUNNING) {
			this.protocol.quit(u);
		}
	}

	/**
	 * Finds a {@link PseudoClient}.
	 *
	 * @param numeric the numeric to look for
	 * @return the PseudoClient in question or null if not found
	 * @since 0.5.0
	 */
	public PseudoClient findPseudoClient(final String numeric)
	{
		return this.pseudoclients_by_id.get(numeric);
	}

	/**
	 * @return a list of all currently known {@link PseudoClients}
	 * @since 0.5.0
	 */
	public Collection<PseudoClient> getPseudoClients()
	{
		return this.pseudoclients_by_id.values();
	}

	/**
	 * Adds a {@link Channel} to the list.
	 *
	 * @param c the Channel to add
	 * @since 0.5.0
	 */
	public void addChannel(final Channel c)
	{
		this.channels.put(protocol.stringToLowerCase(c.getName()), c);
	}

	/**
	 * Finds a {@link Channel} by name.
	 *
	 * @param name the name of the Channel
	 * @return the Channel looked for or null if not found
	 * @since 0.5.0
	 */
	public Channel findChannel(final String name)
	{
		return this.channels.get(protocol.stringToLowerCase(name));
	}

	/**
	 * Removes a {@link Channel} from the list.
	 *
	 * @param c the Channel to remove
	 */
	public void removeChannel(final Channel c)
	{
		c.destroy();
		this.channels.remove(this.protocol.stringToLowerCase(c.getName()));
	}
}
