package invalid.thexbot.irc;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import invalid.thexbot.Util;
import invalid.thexbot.exceptions.ResourceException;
import invalid.thexbot.irc.events.Command;
import invalid.thexbot.protocol.Protocol;

/**
 * Defines a pseudo-client we can use to communicate with users.
 *
 * @since 0.5.0
 */
public final class PseudoClient extends User {
	private final Map<String, Command> commands = new HashMap<>();
	private final Protocol protocol;

	/**
	 * Creates a new pseudo-client. You should not invoke this function directly as it doesn't
	 * add the client to various lists.
	 *
	 * @param p the {@link Protocol} instance X uses
	 * @param nick the nick for the new client
	 * @param ident the ident for the new client
	 * @param host the host for the new client
	 * @param gecos the GECOS (also known as ircname/realnem) for the new client
	 * @throws ResourceException if we are out of user numerics
	 * @since 0.5.0
	 */
	public PseudoClient(Protocol p, String nick, String ident, String host, String gecos) throws ResourceException
	{
		super(p.generateNumeric(), nick, Util.getTimestamp(), ident, ident, "0.0.0.0", host,
			host, gecos, p.getServer());
		this.protocol = p;
		this.setMode(UserMode.OPER);
		this.setMode(UserMode.ADMIN);
		this.setMode(UserMode.SERVICE);
		this.setMode(UserMode.INVISIBLE);
	}

	/**
	 * Adds a {@link Command} to this client.
	 *
	 * @param cmd the Command to add
	 * @since 0.5.0
	 */
	public void addCommand(final Command cmd)
	{
		this.commands.put(cmd.getName().toUpperCase(), cmd);
	}

	/**
	 * Finds a {@link Command}.
	 *
	 * @param name the command name to search for
	 * @return the Command in question or null if not found
	 * @since 0.5.0
	 */
	public Command findCommand(final String name)
	{
		return this.commands.get(name.toUpperCase());
	}
	
	/**
	 * @return all {@link Command}s
	 */
	public Collection<Command> getCommands()
	{
		return this.commands.values();
	}
	
	/**
	 * Removes a {@link Command}.
	 * 
	 * @param name the command to remove
	 * @since 0.5.0
	 */
	public void removeCommand(final String name)
	{
		this.commands.remove(name);
	}
	
	/**
	 * Sends a message to a {@link User} from this client.
	 * @param u the User to talk to
	 * @param message the message to send
	 */
	public void notice(final User u, final String message)
	{
		this.protocol.sendNotice(this, u, message);
	}
	
	@Override
	public void quit()
	{
		super.quit();
		this.protocol.quit(this);
	}
}
