package invalid.thexbot.irc;

import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * Describes an IRC user. Note that pseudo-clients inherit from this class.
 *
 * @since 0.5.0
 */
public class User {
	private final String numeric;
	private String nick;
	private int nickTS;
	private final String real_ident;
	private String shown_ident;
	private final String ip;
	private String real_host; // mutable, as per ircd-hybrid/ratbox sending that late
	private String shown_host;
	private String gecos;
	private Server server;
	// channel
	private final Set<Channel> channels = new HashSet<>();
	/**
	 * Contains metadata. Currently known keys:
	 * <table>
	 * <tr><th>key</th><th>value</th></tr>
	 * <tr><td>account</td><td>AccountInfo</td></tr>
	 * <tr><td>certfp</td><td>String</td></tr>
	 * <tr><td>opername</td><td>String</td></tr>
	 * </table>
	 *
	 * @since 0.5.0
	 */
	private final Map<String, Object> metadata = new HashMap<>();
	private final EnumSet<UserMode> modes = EnumSet.noneOf(UserMode.class);

	/**
	 * Creates a new user.
	 *
	 * @param numeric the user's unique identifier on the network
	 * @param nick the user's nick
	 * @param nickTS the UNIX timestamp when the user got the nick (-1 if not supported by IRCd)
	 * @param real_ident the real ident
	 * @param shown_ident the shown ident (if null, real ident will be filled in)
	 * @param ip the textual representation of the user's source IP address
	 * @param real_host the user's real host
	 * @param shown_host the user's currently shown host (if null, real_host will be filled in)
	 * @param gecos the user's gecos (also known as realname/ircname)
	 * @param server the {@link Server} the user is connected to
	 * @since 0.5.0
	 */
	public User(String numeric, String nick, int nickTS, String real_ident, String shown_ident,
		String ip, String real_host, String shown_host, String gecos, Server server)
	{
		this.numeric = numeric;
		this.nick = nick;
		this.nickTS = nickTS;
		this.real_ident = real_ident;
		this.shown_ident = (shown_ident == null ? this.real_ident : shown_ident);
		this.ip = ip;
		this.real_host = real_host;
		this.shown_host = (shown_host == null ? this.real_host : shown_host);
		this.gecos = gecos;
		this.server = server;
	}

	/**
	 * @return the numeric
	 * @since 0.5.0
	 */
	public String getNumeric()
	{
		return numeric;
	}

	/**
	 * @return the nick
	 * @since 0.5.0
	 */
	public String getNick()
	{
		return nick;
	}

	/**
	 * @param nick the nick to set
	 * @since 0.5.0
	 */
	public void setNick(String nick)
	{
		this.nick = nick;
	}

	/**
	 * @return the nickTS
	 * @since 0.5.0
	 */
	public int getNickTS()
	{
		return nickTS;
	}

	/**
	 * @param nickTS the nickTS to set
	 * @since 0.5.0
	 */
	public void setNickTS(int nickTS)
	{
		this.nickTS = nickTS;
	}

	/**
	 * @return the real ident
	 * @since 0.5.0
	 */
	public String getRealIdent()
	{
		return real_ident;
	}

	/**
	 * @return the shown ident
	 * @since 0.5.0
	 */
	public String getShownIdent()
	{
		return shown_ident;
	}

	/**
	 * @param shown_ident the shown ident to set
	 * @since 0.5.0
	 */
	public void setShownIdent(String shown_ident)
	{
		this.shown_ident = shown_ident;
	}

	/**
	 * @return the textual representation of the user's IP address
	 * @since 0.5.0
	 */
	public String getIP()
	{
		return ip;
	}

	/**
	 * @return the user's real host
	 * @since 0.5.0
	 */
	public String getRealHost()
	{
		return real_host;
	}

	/**
	 * @return the user's shown host
	 * @since 0.5.0
	 */
	public String getShownHost()
	{
		return shown_host;
	}

	/**
	 * @param shown_host the shown host to set
	 * @since 0.5.0
	 */
	public void setShownHost(String shown_host)
	{
		this.shown_host = shown_host;
	}

	/**
	 * @return the GECOS
	 * @since 0.5.0
	 */
	public String getGECOS()
	{
		return gecos;
	}

	/**
	 * @param gecos the GECOS to set
	 * @since 0.5.0
	 */
	public void setGECOS(String gecos)
	{
		this.gecos = gecos;
	}

	/**
	 * Returns a piece of meta data.
	 *
	 * @param key the key to look for
	 * @return the meta data in question or null
	 * @see User#metadata
	 * @since 0.5.0
	 */
	public Object getMetaData(final String key)
	{
		return this.metadata.get(key);
	}

	/**
	 * Adds a piece of meta data.
	 *
	 * @param key the key with which the data is associated
	 * @param data the data to add
	 * @since 0.5.0
	 */
	public void setMetaData(final String key, final Object data)
	{
		this.metadata.put(key, data);
	}

	/**
	 * Removes a piece of meta data.
	 *
	 * @param key the key to clear
	 * @since 0.5.0
	 */
	public void removeMetaData(final String key)
	{
		this.metadata.remove(key);
	}

	/**
	 * Shorthand check whether a user is authenticated.
	 *
	 * @return true if the user is authenticated, false otherwise
	 * @since 0.5.0
	 */
	public boolean isAuthed()
	{
		return this.metadata.get("account") != null;
	}

	/**
	 * @return the user's account info (or null if not authenticated)
	 * @since 0.5.0
	 */
	public AccountInfo getAccountInfo()
	{
		return (AccountInfo)this.metadata.get("account");
	}

	/**
	 * @return the channel list of this user
	 * @since 0.5.0
	 */
	public Set<Channel> getChannels()
	{
		return this.channels;
	}

	/**
	 * Checks if this user has a certain mode.
	 *
	 * @param m the mode to check
	 * @return true if the user has the mode, false otherwise
	 * @since 0.5.0
	 */
	public boolean hasMode(UserMode m)
	{
		return modes.contains(m);
	}

	/**
	 * Adds the given user mode to the user.
	 *
	 * @param m the mode to add
	 * @since 0.5.0
	 */
	public void setMode(UserMode m)
	{
		modes.add(m);
	}

	/**
	 * Removes the given user mode from the user.
	 *
	 * @param m the mode to remove
	 * @since 0.5.0
	 */
	public void unsetMode(UserMode m)
	{
		modes.remove(m);
	}
	
	/**
	 * @return the list of {@link UserMode}s
	 */
	public EnumSet<UserMode> getModes()
	{
		return this.modes;
	}

	/**
	 * @return the {@link Server} this User is connected to
	 * @since 0.5.0
	 */
	public Server getServer()
	{
		return this.server;
	}

	/**
	 * Adds a {@link Channel} to this user. Note that this does neither propagate the change nor
	 * does this change the state of the channel.
	 *
	 * @param c the Channel to add
	 * @since 0.5.0
	 */
	public void join(Channel c)
	{
		this.channels.add(c);
	}

	/**
	 * Removes a {@link Channel} to this user. Note that this does neither propagate the change
	 * nor does this change the state of the channel.
	 *
	 * @param c the Channel to remove
	 * @since 0.5.0
	 */
	public void part(Channel c)
	{
		for (Iterator<Channel> i = this.channels.iterator(); i.hasNext();) {
			if (i.next() == c) {
				i.remove();
			}
		}
	}

	/**
	 * Removes all {@link Channel}s from this User. Does not update Channel state.
	 *
	 * @since 0.5.0
	 */
	public void partAll()
	{
		this.channels.clear();
	}

	/**
	 * Clears lists that this User references. Note that this, too, does not alter Channel
	 * state, which must be cleaned before calling this routine.
	 *
	 * @since 0.5.0
	 */
	public void quit()
	{
		this.server = null;
		this.channels.clear();
		this.metadata.clear();
		this.modes.clear();
	}
}
