package invalid.thexbot.net;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.Socket;
import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Level;
import invalid.thexbot.Logger;

/**
 * Wrapper class that allows for polling lines asynchronously.
 *
 * @since 0.5.0
 */
public class LinePoller {
	private final Socket sock;
	private final BufferedReader reader;
	private final Queue<String> queue = new LinkedBlockingQueue<>();
	private final PollThread poller;

	/**
	 * Initializes a line poller for the given socket.
	 *
	 * @param s socket to pull lines from (not null)
	 * @throws IOException if an IOException was thrown when creating the reader.
	 * @since 0.5.0
	 */
	public LinePoller(Socket sock) throws IOException
	{
		this.sock = sock; // also lock for shutting_down
		this.reader = new BufferedReader(new InputStreamReader(this.sock.getInputStream()));

		this.poller = new PollThread();
	}

	/**
	 * Starts this line poller.
	 *
	 * @since 0.5.0
	 */
	public void start()
	{
		this.poller.start();
	}

	/**
	 * Closes the reader. This does not, however, close the socket.
	 *
	 * @since 0.5.0
	 */
	public void disconnect()
	{
		try {
			this.poller.interrupt();
			this.poller.join();
		} catch (InterruptedException e) {
		}

		try {
			this.reader.close();
		} catch (IOException e) {
		}
	}

	/**
	 * Reads the last line in the line queue.
	 *
	 * @return the oldest line in the queue or null if none available.
	 * @since 0.5.0
	 */
	public String poll()
	{
		synchronized (this.queue) {
			return this.queue.poll();
		}
	}

	private class PollThread extends Thread {
		@Override
		public void run()
		{
			try {
				String line;

				while (true) {
					synchronized (sock) {
						line = reader.readLine();
					}

					if (line == null) {
						Logger.getLogger(LinePoller.class.getName()).log(Level.SEVERE,
							"Received null from reader.");
						break;
					}

					/* NOTE Cannot apply Normalizer here because IRC channels have no encoding and
					 * are thus simply blobs of bytes.
					 */

					synchronized (queue) {
						queue.add(line);
					}

					Logger.getLogger(LinePoller.class.getName()).log(Level.FINEST,
						"Received line " + line);
				}
			} catch (IOException e) {
				Logger.getLogger(LinePoller.class.getName()).log(Level.SEVERE,
					"Exception while polling lines.", e);
			}
		}
	}
}
