package invalid.thexbot.protocol;

import invalid.thexbot.irc.Server;
import invalid.thexbot.irc.User;

/**
 * A MessageHandler receives one message, handles it appropriately and calls events as needed. It
 * also performs validation tasks.
 *
 * @since 0.5.0
 */
public abstract class MessageHandler {
	/**
	 * Describes the type of command sources.
	 *
	 * @since 0.5.0
	 */
	protected enum SourceType {
		/**
		 * Source must be a server, and a server only.
		 *
		 * @since 0.5.0
		 */
		SERVER,
		/**
		 * Source must be a user, and a user only.
		 *
		 * @since 0.5.0
		 */
		USER,
		/**
		 * Source can be anything.
		 *
		 * @since 0.5.0
		 */
		ANY,
		/**
		 * Message has no source.
		 *
		 * @since 0.5.0
		 */
		NONE;
	}
	private final int min_argument_count;
	private final SourceType source_type;
	protected final Protocol protocol;

	/**
	 * Initializes a new message handler.
	 *
	 * @param min_argument_count the minimum arguments to the message
	 * @param source_type the {@link SourceType} this handler accepts
	 * @since 0.5.0
	 */
	public MessageHandler(final Protocol protocol, final int min_argument_count,
		final SourceType source_type)
	{
		this.protocol = protocol;
		this.min_argument_count = min_argument_count;
		this.source_type = source_type;
	}

	/**
	 * @return the minimum amount of arguments this handler requires
	 * @since 0.5.0
	 */
	public int getMinimumArgumentCount()
	{
		return min_argument_count;
	}

	/**
	 * @return the {@link SourceType} this handler requires
	 * @since 0.5.0
	 */
	public SourceType getSourceType()
	{
		return source_type;
	}

	/**
	 * Handles the message with {@link User} source.
	 *
	 * @param m the {@link Message} to handle
	 * @since 0.5.0
	 */
	public abstract void handle(User u, Message m);

	/**
	 * Handles the message with {@link Server} source.
	 *
	 * @param m the {@link Message} to handle
	 * @since 0.5.0
	 */
	public abstract void handle(Server s, Message m);

	/**
	 * Handles the message without source.
	 *
	 * @param m the {@link Message} to handle
	 * @since 0.5.0
	 */
	public abstract void handle(Message m);
}
