package invalid.thexbot.protocol;

import invalid.thexbot.exceptions.ParserException;
import invalid.thexbot.exceptions.ResourceException;
import invalid.thexbot.irc.Channel;
import invalid.thexbot.irc.PseudoClient;
import invalid.thexbot.irc.Server;
import invalid.thexbot.irc.User;

/**
 * Describes ways to interface with the uplink.
 *
 * @since 0.5.0
 */
public interface Protocol {
	/**
	 * @return the {@link Server} that describes this pseudo-server
	 */
	public Server getServer();

	/**
	 * @return an unused numeric for a pseudo-client
	 * @throws ResourceException if there are no more user numerics available
	 */
	public String generateNumeric() throws ResourceException;

	/**
	 * Initiates the connection as needed.
	 *
	 * @since 0.5.0
	 */
	public void handshake();

	/**
	 * Sends our information.
	 *
	 * @since 0.5.0
	 */
	public void burst();

	/**
	 * Acknowledges the end of burst from the uplink.
	 *
	 * @since 0.5.0
	 */
	public void acknowledgeEndOfBurst();

	/**
	 * Sends PONG to a {@link Server}.
	 *
	 * @param s the Server to talk to
	 * @since 0.5.0
	 */
	public void pong(Server s);

	/**
	 * Sends PONG to a {@link User}.
	 *
	 * @param u the User to respond to
	 * @since 0.5.0
	 */
	public void pong(User u);

	/**
	 * Sends the end-of-burst signal for us to the uplink.
	 *
	 * @since 0.5.0
	 */
	public void endOfBurst();

	/**
	 * Disconnects from the IRC server. This does not close the socket itself.
	 *
	 * @since 0.5.0
	 */
	public void disconnect();

	/**
	 * Sends a numeric.
	 *
	 * @param num the number to send
	 * @param message the message to send
	 * @since 0.5.0
	 */
	public void numeric(int num, String raw_message);

	/**
	 * Introduces a {@link PseudoClient}.
	 *
	 * @param u the client to introduce
	 * @since 0.5.0
	 */
	public void introducePseudoClient(PseudoClient u);

	/**
	 * Sends a privmsg to a {@link User}. This method may split the message up into multiple
	 * parts if it exceeds the 512 character limit. The message will first be split up by "\n".
	 *
	 * @param source the {@link PseudoClient} to send the message from; if null, the server will
	 * be used
	 * @param u the target User
	 * @param message the message to send
	 * @since 0.5.0
	 */
	public void sendPrivmsg(PseudoClient source, User u, String message);

	/**
	 * Sends a notice to a {@link User}. This method may split the message up into multiple
	 * parts if it exceeds the 512 character limit. The message will first be split up by "\n".
	 *
	 * @param source the {@link PseudoClient} to send the message from; if null, the server will
	 * be used
	 * @param u the target User
	 * @param message the message to send
	 * @since 0.5.0
	 */
	public void sendNotice(PseudoClient source, User u, String message);

	/**
	 * Propagates a {@link PseudoClient} quitting.
	 *
	 * @param u the client to quit
	 */
	public void quit(PseudoClient u);

	/**
	 * Parses a line into a list of {@link Message}s.
	 *
	 * @param line the line to be parsed
	 * @throws ParserException if the line could not be cleanly parsed
	 * @return one or more messages, or null if line was null
	 * @since 0.5.0
	 */
	public Message parse(String line) throws ParserException;

	/**
	 * Parses the given user modes (with arguments, if needed). This will directly modify the
	 * given user.
	 *
	 * @param u the {@link User} the modes apply to
	 * @param modearray the modes with arguments to be parsed
	 * @since 0.5.0
	 */
	public void parseUserModes(User u, String[] modearray);

	/**
	 * Returns the String representation of a user mode set.
	 *
	 * @param u the user whose modes are to be stringified
	 * @return the String, which may be "+" but not null
	 * @since 0.5.0
	 */
	public String getUserModeString(User u);

	/**
	 * Changes the {@link User}'s host to be the hidden host if necessary. Note that protocols
	 * using UnrealIRCd-style cloaking (which is essentially only the P10 family) should dummy
	 * this function out because +x information is sent over the network as shown hosts anyway.
	 *
	 * @param u the User whose host is to be changed
	 * @since 0.5.0
	 */
	public void hideUserHost(User u);

	/**
	 * Parses the given channel modes (with arguments, if needed). This will directly modify the
	 * given {@link Channel}
	 *
	 * @param c the Channel the modes apply to
	 * @param modearray the modes with arguments to be parsed
	 * @since 0.5.0
	 */
	public void parseChannelModes(Channel c, String[] modearray);

	/**
	 * Causes the protocol to actually handle the given message. This will cause the protocol
	 * module to update the internal state, too.
	 *
	 * @param message the {@link Message} to handle
	 * @since 0.5.0
	 */
	public void interpret(Message message);

	/**
	 * Checks if two Strings are equal according to the upper/lowercase rules of the protocol.
	 *
	 * @param str1 the first String
	 * @param str2 the second String
	 * @return true if they are equal, in any other case or one of the strings was null unless
	 * both are null
	 * @since 0.5.0
	 */
	public boolean stringsEqualIgnoreCase(String str1, String str2);

	/**
	 * Converts a string to lower-case according to the upper/lowercase rules of the protocol.
	 *
	 * @param s the String to convert
	 * @return the converted String or null if s was null
	 * @since 0.5.0
	 */
	public String stringToLowerCase(String s);
}
