package invalid.thexbot.protocol;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import invalid.thexbot.exceptions.ParserException;
import invalid.thexbot.irc.events.EventHandler;

abstract class ProtocolBase implements Protocol {
	//private final Socket sock;
	private final PrintWriter out;
	//private static final Logger log = Logger.getLogger(ProtocolBase.class.getName());
	protected final Set<EventHandler> event_handlers = new HashSet<>();

	public ProtocolBase(final Socket sock) throws IOException
	{
		//this.sock = sock;
		this.out = new PrintWriter(new OutputStreamWriter(sock.getOutputStream()), true);
	}

	/**
	 * Flushes the outgoing buffer and closes the socket.
	 */
	@Override
	public void disconnect()
	{
		this.out.close();
	}

	/**
	 * Sends a line to the uplink.
	 *
	 * @param line the line to be sent; do not append \r\n
	 */
	public void sendLine(final String line)
	{
		this.out.println(line);
	}

	/**
	 * Sends a line formatted with {@link String#format(java.lang.String, java.lang.Object[])}
	 *
	 * @param format the string format
	 * @param args the arguments for the format
	 */
	public void sendFormattedLine(final String format, final Object... args)
	{
		sendLine(String.format(format, args));
	}

	/**
	 * Parses an incoming line based on the RFC1459 specification.
	 *
	 * @param line the line to parse
	 * @return a parsed {@link Message}
	 */
	@Override
	public Message parse(String line) throws ParserException
	{
		if (line == null) {
			return null;
		}

		char[] buf = line.toCharArray();
		int colonat = line.indexOf(':', 2);
		int spaces = 0;

		// In other words, have we gotten a source parameter?
		if (colonat != -1) {
			for (int i = 0; i < colonat; i++) {
				if (buf[i] == ' ') {
					spaces++;
				}
			}
		} else {
			// Don't split after the colon. To do so, keep track of the spaces.
			for (int i = 0; i < buf.length; i++) {
				if (buf[i] == ' ') {
					spaces++;
				}
			}
		}

		String[] pieces = line.split(" ", spaces + 1);
		boolean has_source = pieces[0].charAt(0) == ':';
		if (pieces[pieces.length - 1].startsWith(":")) {
			pieces[pieces.length - 1] = pieces[pieces.length - 1].substring(1);
		}

		String source = has_source ? pieces[0].substring(1) : null;
		String cmd = has_source ? pieces[1] : pieces[0];
		String[] cmdargs = Arrays.copyOfRange(pieces, has_source ? 2 : 1, pieces.length);

		return new Message(source, cmd, cmdargs);
	}

	/**
	 * Compares equality of two Strings by RFC1459 rules.
	 *
	 * @param str1 one of the Strings to compare
	 * @param str2 the other of the Strings to compare
	 * @return true if str1 == str2, they are equal according to the RFC1459 casing rules, false
	 * otherwise.
	 */
	@Override
	public boolean stringsEqualIgnoreCase(final String str1, final String str2)
	{
		if (str1 == null || str2 == null) {
			if (str1 == null && str2 == null) {
				return true;
			}
			return false;
		}

		if (str1 == str2) {
			return true;
		}

		if (str1.length() != str2.length()) {
			return false;
		}

		return stringToLowerCase(str1).equals(stringToLowerCase(str2));
	}

	/**
	 * Converts a String to lower-case according to RFC1459 casing rules.
	 *
	 * @param s the String to convert
	 * @return the converted String
	 */
	@Override
	public String stringToLowerCase(String s)
	{
		char[] buf = s.toCharArray();

		for (int i = 0; i < buf.length; i++) {
			if (buf[i] == '[') {
				buf[i] = '{';
			} else if (buf[i] == ']') {
				buf[i] = '}';
			} else if (buf[i] == '\\') {
				buf[i] = '|';
			} else {
				buf[i] = Character.toLowerCase(buf[i]);
			}
		}

		return new String(buf);
	}
}
