package invalid.thexbot.scripting;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import invalid.thexbot.Logger;
import invalid.thexbot.TheXBot;
import invalid.thexbot.exceptions.ModuleException;
import invalid.thexbot.irc.Source;
import invalid.thexbot.irc.events.EventHandler;
import invalid.thexbot.protocol.Protocol;
import org.python.util.PythonInterpreter;

/**
 * Keeps track of modules.
 *
 * @since 0.5.0
 */
public final class ModuleManager {
	private static final Logger log = Logger.getLogger(ModuleManager.class.getName());
	private final Protocol protocol;
	private final Map<String, Module> modules = new HashMap<>();
	private final PythonInterpreter interpreter;

	/**
	 * Initializes this module manager.
	 *
	 * @param protocol the {@link Protocol} instance we talk through
	 * @since 0.5.0
	 */
	public ModuleManager(Protocol protocol)
	{
		this.protocol = protocol;
		this.interpreter = new PythonInterpreter();
	}

	/**
	 * Loads the modules as specified by the configuration file.
	 *
	 * @since 0.5.0
	 */
	public void loadModules()
	{
		for (final String name : TheXBot.conf.getModuleNames()) {
			try {
				loadModule(name);
			} catch (Exception e) {
				log.log(Level.SEVERE, "Error loading module ", e);
			}
		}
	}

	/**
	 * Loads a module.
	 *
	 * @param name the name of the module
	 * @return the {@link Module} that was loaded
	 * @since 0.5.0
	 */
	public Module loadModule(final String name) throws Exception
	{
		if (this.modules.get(name) != null) {
			throw new ModuleException("Module " + name + " is already loaded!");
		}
		
		log.log(Level.INFO, "Loading module " + name + "...");
		
		this.interpreter.set("protocol", this.protocol);
		this.interpreter.set("jlogger", Logger.getLogger("Module$" + name));
		this.interpreter.execfile("scripts/" + name + ".py");
		
		Module module = this.interpreter.get("__module__", Module.class);
		this.modules.put(name, module);
		log.log(Level.INFO, "Module " + name + " loaded.");

		return module;
	}

	/**
	 * Unloads a module.
	 *
	 * @param name the name of the module to unload
	 * @return true if the module was loaded, false if it wasn't
	 * @since 0.5.0
	 */
	public boolean unloadModule(final String name)
	{
		log.log(Level.INFO, "Unloading module " + name + "...");
		Module m = this.modules.get(name);
		if (m == null) {
			return false;
		}

		m.__deinit__();
		this.modules.remove(name);
		log.log(Level.INFO, "Module " + name + " unloaded.");
		return true;
	}

	/**
	 * Unloads all modules and stops the Python interpreter.
	 *
	 * @since 0.5.0
	 */
	public void cleanup()
	{
		log.log(Level.INFO, "Unloading ALL modules.");
		for (final Module m : this.modules.values()) {
			m.__deinit__();
		}
		this.modules.clear();
		this.interpreter.cleanup();
	}

	/**
	 * Runs an event.
	 *
	 * @param name the event name, this must be the same as the function name in
	 * {@link EventHandler}
	 * @param source the source of the command (may be null)
	 * @param args various arguments to the event as needed
	 * @since 0.5.0
	 */
	public void runEvent(String name, Source source, Object[] args)
	{
		for (final Module module : this.modules.values()) {
			try {
				Method m = module.getClass().getDeclaredMethod(name, Source.class, Object[].class);
				m.invoke(module, source, args);
			} catch (NoSuchMethodException e) {
				// ignore because modules don't have to implement those
			} catch (SecurityException | IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
				Logger.getLogger(ModuleManager.class.getName()).log(Level.SEVERE, "Error while running events.", e);
			}
		}
	}
}
